﻿------------------------------------------------------------------------------
-- $Id$
------------------------------------------------------------------------------
--                           module_destruction.lua
--
--	Copyright (C) 2012 Deronda, Maighurl
--
--  This file is for use with clcInfo by abija.  Some of the code was pulled
--  from existing clcInfo modules and the example module.
--
--  abija has distributed clcInfo All Rights Reserved.  For more information
--  go to: http://wow.curseforge.com/addons/clctracker/
--
------------------------------------------------------------------------------
-- Author:
--      Deronda (@ Stormreaver), Maighurl
------------------------------------------------------------------------------
-- Description:
--   Actions module for destruction warlocks
------------------------------------------------------------------------------
-- make sure we should be loading this
if not clcInfo then return end       -- need clcInfo to run
if not clcInfo_Warlock then return end -- need clcInfo_Warlock
local _, englishClass = UnitClass("player") -- need to be a Warlock
if englishClass ~= clcInfo_Warlock.CLASS then return end

-- Get local copies of functions/tables.  This will speed up execution.
local clcInfo_Warlock = clcInfo_Warlock
local spells = clcInfo_Warlock.spells
local GetTimeLeft = clcInfo_Warlock.GetTimeLeft
local GetCooldown = clcInfo_Warlock.GetCooldown
local IconNext = clcInfo_Warlock.IconNext
local IconCurrent = clcInfo_Warlock.IconCurrent
local CLIP_HARMFUL = clcInfo_Warlock.CLIP_HARMFUL
local UnitPower = UnitPower

local guess_travel_time_immolate = clcInfo_Warlock.spells["Immolate"].GuessTravelTime
local guess_travel_time_soul_fire = clcInfo_Warlock.spells["Soul Fire"].GuessTravelTime
local is_in_flight_immolate = clcInfo_Warlock.spells["Immolate"].IsInFlight
local is_in_flight_soul_fire = clcInfo_Warlock.spells["Soul Fire"].IsInFlight
local get_time_left_demon_soul = clcInfo_Warlock.GetTimeLeftDemonSoul
local calculate_target_health_pct = clcInfo_Warlock.CalculateTargetHealthPct

-- Assign module variables
local specname = "destruction"
local prio_arg = "destructionprio"
local sample_prio = "soulburn ds sb_sf sf_isf immolate conflagrate bod corruption shadowflame chaosbolt eimp_sf shadowburn incinerate lifetap"
local priority = {"soulburn","none","sb_sf", "sf_isf", "immolate","conflagrate","bod","corruption",
                  "shadowflame","chaosbolt","eimp_sf","shadowburn","incinerate","lifetap"}
local class_module
local class_module_name

-- user variables
-- character specific variables.  Maybe one day we'll use these.
local db_default = {
    char_var = false,
}
local db

-- template specific variables.
local tdb_default = {
    version = clcInfo_Warlock.VERSION,
    priority = priority,
}
local tdb

-- actions -------------------------------------------------------------------------
-- local copies of spell names and ids to ugly up code but speed action functions up
-- we want the action functions to run as quickly as possible
local Backdraft = spells["Backdraft"].name
local BaneofAgony = spells["Bane of Agony"].name
local BaneofDoom = spells["Bane of Doom"].name
local ChaosBolt = spells["Chaos Bolt"].name
local Conflagrate = spells["Conflagrate"].name
local Corruption = spells["Corruption"].name
local DemonSoul = spells["Demon Soul"].name
local EmpoweredImp = spells["Empowered Imp"].name
local Immolate = spells["Immolate"].name
local ImprovedSoulFire = spells["Improved Soul Fire"].name
local Incinerate = spells["Incinerate"].name
local Soulburn = spells["Soulburn"].name
local SoulFire = spells["Soul Fire"].name

local BackdraftId = spells["Backdraft"].id
local BaneofDoomId = spells["Bane of Doom"].id
local ChaosBoltId = spells["Chaos Bolt"].id
local ConflagrateId = spells["Conflagrate"].id
local CorruptionId = spells["Corruption"].id
local DemonSoulId = spells["Demon Soul"].id
local DoomguardId = spells["Summon Doomguard"].id
local ImmolateId = spells["Immolate"].id
local IncinerateId = spells["Incinerate"].id
local LifeTapId = spells["Life Tap"].id
local ShadowburnId = spells["Shadowburn"].id
local ShadowflameId = spells["Shadowflame"].id
local SoulburnId = spells["Soulburn"].id
local SoulFireId = spells["Soul Fire"].id

-- A little function to return the casting time multiplier for
-- Chaos Bolt and Incinerate.
local function get_backdraft_multiplier(casting_name, cur_id)
    local multiplier = 1 -- status quo
    local name, _, _, count = UnitAura("player", Backdraft, nil, "HELPFUL")

    if name then
        -- if we have backdraft casting time checks for Chaos Bolt and
        -- Incinerate will be 30% lower.  We need to see if it will drop.
        if casting_name ~= Conflagrate and cur_id ~= ConflagrateId then
            if casting_name == ChaosBolt or casting_name == Incinerate then
                count = count - 1
            end
            if cur_id == ChaosBoltId or cur_id == IncinerateId then
                count = count - 1
            end
            if count < 1 or GetTimeLeft("player", Backdraft, "HELPFUL") <= 0 then
                multiplier = 1/0.7 -- backdraft will drop, remove 30% buff
            end
        end
    else
        -- if backdraft isn't up, casting time checks for Chaos Bolt and
        -- Incinerate will be normal.  We need to see if it will pop up.
        if casting_name == Conflagrate or cur_id == ConflagrateId then
            multiplier = 0.7 --Backdraft will proc, drop casting time by 30%
        end
    end
    
    return multiplier
end

-- ACTIONS!!!!!
-- The action table contains an entry for each priority.  A priority is also
-- a table that has three entries:
--    1) id (number)-> the spell id to return for the action
--    2) GetCD() -> a function that, when run, returns the cooldown remaining
--                  for the action.
--    3) info (string) -> the localized name of the action.  This could be
--               the spell name, such as "Immolate", or a description of the
--               situation, for instance, "Empowered Imp: Soul Fire".
--
-- Writing the GetCD() function can be tricky, but therein lies the power of
-- this code.  Remember that the GetCD() function will be run twice per action.
-- Once to suggest the current spell and a second time to suggest the next spell.
-- It must handle both cases.
--
-- This isn't as difficult as it sounds.  The functions GetTimeLeft() and
-- GetCooldown() already take into account the state variables stored in
-- clcInfo_Warlock.lua under the hood.  For instance, a GetTimeLeft() called
-- on the Immolate DoT might be 13.5 during the first call to an action's GetCD().
-- If Bane of Doom is suggested for the current ability, a call to GetTimeLeft()
-- during the second call to GetCD() would return 12 since Bane has a cast time
-- of 1.5 seconds (assuming no haste buff).  The arguments for GetCD() are:
--
-- seconds, multiplier = GetCD(casting_name, cur_id, mana, distance, gcd_duration)
--
-- seconds - the number of seconds before this spell should be cast.
-- multiplier - a multiplier for WarlockRotation() to use on casting time if
--              this action is selected for the cur_id.  Usually should be 1.
--              For instant cast spells not on the GCD like Demon Soul set to 0.
-- casting_name - the name of the spell being channelled by the player or
--                nil if not channelling.
-- cur_id - nil if the current id hasn't been determined yet (in other words,
--          this call is determining it!).  The current spell id of the spell
--          to display in the current ability icon otherwise.
-- mana - how much mana the player has.
-- distance - the distance to the target.  This is the max range variable from
--            GetRange().
-- gcd_duration - how long a gcd takes.
--
-- see below for examples
--
local actions = {
	soulburn = {
		id = SoulburnId,
		GetCD = function(_, cur_id)
            -- we can't cast this twice in a row, we can't cast this without
            -- shards, and we won't cast it if Soulburn is actually up
            if cur_id ~= SoulburnId
               and UnitPower("player", SPELL_POWER_SOUL_SHARDS) > 0
               and GetTimeLeft("player", Soulburn, "HELPFUL") == 0 then
                return GetCooldown(SoulburnId), 0
            end
            return 100
        end,
		info = spells["Soulburn"].name,
	},
	ds = {
		id = DemonSoulId,
		GetCD = function(_, cur_id, mana)
	        if cur_id ~= DemonSoulId            -- can't cast this twice in a row
               and UnitExists("playerpet") then -- can't cast without a pet
                local _, _, _, cost = GetSpellInfo(DemonSoulId)
                if (mana > cost or cur_id == LifeTapId) then -- can't cast without mana
                    return GetCooldown(DemonSoulId), 0
                end
            end
            return 100
		end,
		info = spells["Demon Soul"].name,
	},
    sb_sf = { -- soul fire if there's a soul burn
		id = SoulFireId,
		GetCD = function(_, cur_id, mana)
            if cur_id ~= SoulFireId
               and (cur_id == SoulburnId or GetTimeLeft("player", Soulburn, "HELPFUL") > 0) then
                local _, _, _, cost = GetSpellInfo(SoulFireId)
                if (mana > cost or cur_id == LifeTapId) then
                    return GetCooldown(SoulFireId), 1
                end
            end
            return 100
		end,
		info = spells["Soulburn"].name..":"..spells["Soul Fire"].name,
	},
    sf_isf = { -- keep Improved Soul Flame up
		id = SoulFireId,
		GetCD = function(casting_name, cur_id, mana, distance, gcd_duration)
            -- just for keeping ISF up do not cast twice in a row
            if cur_id ~= SoulFireId
               and casting_name ~= SoulFire
               and not is_in_flight_soul_fire() then
                local _, _, _, cost, _, _, casting_time = GetSpellInfo(SoulFireId)
                if (mana > cost or cur_id == LifeTapId) then
                    -- we want to cast this before the buff drops.  So get the time
                    -- left on the buff MINUS the casting time + travel time + gcd.
                    -- Return that as the cd unless it is less than the actual cd of
                    -- the spell
                    return max(GetTimeLeft("player", ImprovedSoulFire, "HELPFUL")
                                -(casting_time/1000 + guess_travel_time_soul_fire(distance)
                                + gcd_duration),0), 1
                end
            end
            return 100
		end,
		info = spells["Improved Soul Fire"].name..":"..spells["Soul Fire"].name,
	},
    immolate = {
		id = ImmolateId,
		GetCD = function(casting_name, cur_id, mana)
            -- don't cast if already casting or one is on the way
            if cur_id ~= ImmolateId
               and casting_name ~= Immolate
               and not is_in_flight_immolate() then
                local _, _, _, cost, _, _, casting_time = GetSpellInfo(ImmolateId)
                if (mana > cost or cur_id == LifeTapId) then
                    -- Do not cast until the time left on the DoT is low enough.
                    -- Return the time left MINUS the casting time + travel time
                    -- and then subtract the clip time as well.  Return it or the
                    -- spell CD, whichever is greater.
                    return max(GetTimeLeft("target", Immolate, "HARMFUL|PLAYER")
                                - (casting_time/1000 + guess_travel_time_immolate())
                                - CLIP_HARMFUL, 0), 1
                end
            end
            return 100
		end,
		info = spells["Immolate"].name,
	},
    conflagrate = {
		id = ConflagrateId,
		GetCD = function(casting_name, cur_id, mana)
            --since it has a cd, can't cast twice in a row.
            --only suggest if immolate is already being cast, suggested,
            --on the way, or ticking
            if cur_id ~= ConflagrateId then
                local _, _, _, _, _, _, casting_time = GetSpellInfo(ImmolateId)
                if casting_name == Immolate
                   or cur_id == ImmolateId
                   or is_in_flight_immolate()
                   or GetTimeLeft("target", Immolate, "HARMFUL|PLAYER") > 0 then
                    _, _, _, cost = GetSpellInfo(ConflagrateId)
                    if (mana > cost or cur_id == LifeTapId) then
                        return GetCooldown(ConflagrateId), 1
                    end
                end
            end
            return 100
		end,
		info = spells["Conflagrate"].name,
	},
    bod = {
		id = BaneofDoomId,
		GetCD = function(_, cur_id, mana)
            if cur_id ~= BaneofDoomId then
                local _, _, _, cost = GetSpellInfo(BaneofDoomId)
                if (mana > cost or cur_id == LifeTapId) then
                    return max(GetTimeLeft("target", BaneofDoom, "HARMFUL|PLAYER")-CLIP_HARMFUL,
                                 GetTimeLeft("target", BaneofAgony, "HARMFUL|PLAYER")), 1
                end
            end
            return 100
		end,
		info = spells["Bane of Doom"].name,
	},
    corruption = {
		id = CorruptionId,
		GetCD = function(_, cur_id, mana)
            if cur_id ~= CorruptionId then
                local _, _, _, cost = GetSpellInfo(CorruptionId)
                if (mana > cost or cur_id == LifeTapId) then
                    return max(GetTimeLeft("target", Corruption, "HARMFUL|PLAYER")-CLIP_HARMFUL,0), 1
                end
            end
            return 100
		end,
		info = spells["Corruption"].name,
	},
    shadowflame = {
		id = ShadowflameId,
		GetCD = function(_, cur_id, mana, distance)
            -- since distance is max_distance from GetRange, we're looking
            -- for the 8 from the 5,8 pair.
            -- since it has a cd, don't suggest twice
            if cur_id ~= ShadowflameId and distance <= 8 then
                local _, _, _, cost = GetSpellInfo(ShadowflameId)
                if (mana > cost or cur_id == LifeTapId) then
                    return GetCooldown(ShadowflameId), 1
                end
            end
            return 100
		end,
		info = spells["Shadowflame"].name,
	},
    chaosbolt = {
		id = ChaosBoltId,
		GetCD = function(casting_name, cur_id, mana)
            --since this has a cd, don't suggest twice
            --also check the casting_time so it's not suggested when
            --the warlock's haste value is too high
            if cur_id ~= ChaosBoltId and casting_name ~= ChaosBolt then
                local _, _, _, cost, _, _, casting_time = GetSpellInfo(ChaosBoltId)
                local multiplier = get_backdraft_multiplier(casting_name, cur_id)
                casting_time = casting_time * multiplier
                if (mana > cost or cur_id == LifeTapId) and casting_time > 900 then
                    return GetCooldown(ChaosBoltId), multiplier
                end
            end
            return 100
		end,
		info = spells["Chaos Bolt"].name,
	},
    eimp_sf = {
		id = SoulFireId,
		GetCD = function(_, cur_id, mana)
            if cur_id ~= SoulFireId
               and GetTimeLeft("player", EmpoweredImp, "HELPFUL") > 0 then
                local _, _, _, cost = GetSpellInfo(SoulFireId)
                if (mana > cost or cur_id == LifeTapId) then
                    return GetCooldown(SoulFireId), 1
                end
            end
            return 100
		end,
		info = spells["Empowered Imp"].name..":"..spells["Soul Fire"].name,
	},
    shadowburn = {
		id = ShadowburnId,
		GetCD = function(_, cur_id, mana)
            -- has a cd so don't suggest twice.  Target must be < 20% health.
            if cur_id ~= ShadowburnId and calculate_target_health_pct() < 20 then
                local _, _, _, cost = GetSpellInfo(ShadowburnId)
                if (mana > cost or cur_id == LifeTapId) then
                    return GetCooldown(ShadowburnId), 1
                end
            end
            return 100
		end,
		info = spells["Shadowburn"].name,
	},
    incinerate = {
		id = IncinerateId,
		GetCD = function(casting_name, cur_id, mana)
            -- CAN be suggested twice in a row, so don't restrict
            -- just check mana
            local _, _, _, cost = GetSpellInfo(ShadowflameId)
            if (mana > cost or cur_id == LifeTapId) then
                return GetCooldown(IncinerateId), get_backdraft_multiplier(casting_name, cur_id)
            end
            return 100
        end,
		info = spells["Incinerate"].name,
	},
    lifetap = {
		id = LifeTapId,
		GetCD = function()
            --we only get here if all other spells don't have mana
            --to cast.  Just return cd
            return GetCooldown(LifeTapId), 1
		end,
		info = spells["Life Tap"].name,
	},
}

----FUNCTIONS-------------------------------------------------------------------
-- The following are called once or infrequently, so speed isn't an issue.
-- This is where the module is created and initialized.

-- called whenever module options are loaded
local GetDB
local SetDB
local GetTDB
local SetTDB
local GetPriority
local SetPriority

--initialize module.
class_module, class_module_name = clcInfo_Warlock:init_module(specname)

--called when template updates.
function class_module:OnTemplatesUpdate()
	tdb, GetTDB, SetTDB, GetPriority, SetPriority
        = clcInfo_Warlock:OnTemplatesUpdate(class_module_name, actions, tdb_default, specname)
end

-- called each time clcInfo initializes
function class_module.OnInitialize()
	db, GetDB, SetDB = clcInfo_Warlock:OnInitialize(class_module_name, db_default)
end

-- called when module options are opened
local function LoadDestructionModule()
    local _, talent_tree_name = GetTalentTabInfo(3) -- localize "Destruction"
    clcInfo_Warlock.LoadModule(class_module_name, talent_tree_name, sample_prio, actions,
                               GetDB, SetDB, GetTDB, SetTDB, GetPriority, SetPriority)
end
clcInfo.optionsCMLoaders[#(clcInfo.optionsCMLoaders) + 1] = LoadDestructionModule

-- giving queue as command line
local function CmdDestruction(args)
	clcInfo_Warlock:CmdLine(tdb, args, actions, specname)
end

clcInfo.cmdList[prio_arg] = CmdDestruction

----ABILITY ICONS--------------------------------------------------------------
----Add these to the clcInfo.env (so they live with IconAura, etc.
local emod = clcInfo.env
function emod.IconDestructionCurrent(...)
	return IconCurrent(actions, specname, tdb.rangePerSkill)
end
function emod.IconDestructionNext(...)
	return IconNext()
end
